package cse378;
import java.util.Arrays;

import cse378.address.PhysicalAddress;
import cse378.cache.CacheLine;
import cse378.status.Status;
import cse378.util.BitField;
import cse378.util.Bits;

// Abstract AbstractCache class

/** Base class for CSE 378 AbstractCache lab 5
*/
public abstract class AbstractCache {
	
	protected CacheLine[] cachelines; //array of cachelines in the cache
	protected int numLines;			//number of cachelines
	protected int lineBits;			//number of bits to index the lines
	protected int numSets; 			//number of entries / line
	protected int blockSize;        //number of words / entry
	protected int blockBits;		//

	protected Status stat;

	protected long hits;
	protected long cleanMisses;
	protected long dirtyMisses;
	
	/** Simple constructor
	 * 
	 * @param numLines number of lines in the cache
	 */
    public AbstractCache(int numLines) {
		this(numLines, 1, 1);
    }
    
    /** Set Associative Constructor
     * 
     * @param numLines number of lines in the cache
     * @param associativity number of entries per line
     */
    public AbstractCache(int numLines, int associativity){
    	this(numLines,associativity,  1);
    }
    
    /** Full Constructor
     * 
     * @param numLines number of lines in the cache
     * @param associativity number of entries per line
     * @param blockSize number of words per entry
     */
    public AbstractCache(int numLines, int associativity, int entrySize){
    	assert(numLines > 0); // Must have at least 1 line
    	assert(32 > associativity && associativity > 0); // Must have at least 1 set 
    	assert(16 > entrySize && entrySize > 0); // Must have at least 1 word / entry
    	
    	this.numLines = numLines;
    	this.lineBits = Bits.log2(numLines);
       	this.numSets = associativity;
    	this.blockSize = entrySize;
       	this.blockBits = Bits.log2(entrySize << 2);
    	this.cachelines = new CacheLine[this.numLines];
    	int tagBits = Bits.log2(1 << (32 - lineBits - blockBits));

    	this.stat = null;
 
    	this.hits = 0;
    	this.cleanMisses = 0;
    	this.dirtyMisses = 0;
    	
    	// populate the cache with CacheLines
    	Arrays.fill(cachelines, new CacheLine(tagBits, this.blockSize, this.numSets));
   }

    public String getInfo(){
    	return "AbstractCache Info: "+this.numLines+" Lines, "+this.numSets+" Way, "+this.blockSize+" Words/AbstractBlock\n";
    }
    
    /** Retrieve status for the most recent probe */
    public Status getStatus(){
    	return stat;
    }
    
    /**load from the cache at location
	*@param addr PhysicalAddress representing the physical location of the load target
	*/
	abstract public void load (PhysicalAddress paddr);
	
	/**store to cache at location
	*@param paddr PhysicalAddress to update
	*
	*/
	abstract public void store (PhysicalAddress paddr);

	/**Compute the line corresponding to a particular location
	*@param location location in memory
	*@return index of the line addressed by location
	*/
	 abstract protected BitField computeIndex(PhysicalAddress location);
	
	/**Compute the tag corresponding to a particular location
	 * @param location location in memory
	 * @return tag portion of location
	 */
	 abstract protected BitField computeTag(PhysicalAddress location);
	
	
	/** Get the line addressed by the index
	 * @param index a BitField specifying the index
	 *
	 * @return AbstractLine addressed by the index
	 */
	 abstract protected AbstractLine getLine(BitField index);
	
	 public long getHits(){
		 return this.hits;
	 }

	 public long getCleanMisses(){
		 return this.cleanMisses;
	 }
	 
	 public long getDirtyMisses(){
		 return this.dirtyMisses;
	 }

	 public String getCounts(){
		 long accesses = this.hits + this.cleanMisses + this.dirtyMisses;
		 return accesses+"\t"+this.hits+"\t"+this.cleanMisses+"\t"+this.dirtyMisses;
	 }
	 
}

