package cse378.pagetable;
import java.util.*;

import cse378.address.PhysicalPage;
import cse378.address.VirtualAddress;
import cse378.address.VirtualPage;
import cse378.status.Status;

import cse378.status.*;


/**
 * @author alaska
 *
 */
public class PageTable {

	HashMap<VirtualPage,PageTableEntry> pages;
	LinkedList<PageTableEntry> recent;

	final int maxPages;
	int pageCount;

	long hits;
	long cleanMisses;
	long dirtyMisses;

	Status stat;

	VirtualPage tlbEvictionAddress;

	/** Create a page table with a fixed number of physical pages
	 * 
	 * @param maxSize number of physical pages available
	 * 
	 */
	public PageTable(int maxSize){
		pages = new HashMap<VirtualPage,PageTableEntry>();
		stat = new Status(Component.PAGETABLE);
		recent = new LinkedList<PageTableEntry>();
		maxPages = maxSize;
		hits = 0;
		dirtyMisses = 0;
		cleanMisses = 0;
	}


	/** Initialize the page table status
	 * 
	 * @param addr address of current access
	 */
	protected void initStatus(VirtualAddress addr){
		stat.setAddress(addr);
		stat.setAccess(Access.INIT);
		tlbEvictionAddress = null;
	}

	/** Get a physical page mapping from the page table
	 *  If none is found, one is created
	 * @param addr VirtualAddress of the desired page
	 * @see AbstractCache#load(PhysicalAddress)
	 */

	public PhysicalPage load(VirtualAddress addr) {
		// reset Status & tlbEvictionAddress
		initStatus(addr);

		VirtualPage pageNum = addr.getPage();
		PhysicalPage pageFrame = null;
		PageTableEntry theEntry = pages.get(pageNum);
		if((theEntry == null) ||
				(theEntry.isSwapped()) ||
				!(theEntry.isValid())){
			if(pageCount < maxPages){
				pageFrame = new PhysicalPage(pageCount++);
				stat.setAccess(Access.CLEANMISS);
			} else {
				PageTableEntry oldEntry = getLRU();
				pageFrame = oldEntry.getPhysical();
				if(oldEntry.isValid() && oldEntry.isDirty()){
					stat.setDirty();
					this.dirtyMisses++;
					oldEntry.setSwapped(true);
				} else {
					this.cleanMisses++;
					stat.setClean();
				}
				oldEntry.setValid(false); // invalidate old entry
				this.tlbEvictionAddress = oldEntry.getTag();
			}
			theEntry = new PageTableEntry(pageNum,pageFrame);
			pages.put(pageNum,theEntry);
		} else {
			this.hits++;
		}
		setMRU(theEntry);
		return theEntry.getPhysical();
	}

	
	
	private PageTableEntry getLRU(){
		return recent.removeFirst();
	}

	public void setMRU(PageTableEntry pte){
		recent.addLast(pte);
	}

	
	public VirtualPage pendingTLBEviction(){
		return tlbEvictionAddress;
	}

	/** Get a physical page mapping from the page table
	 *  If none is found, one is created, (and marked dirty)
	 *  
	 * @param addr VirtualAddress of the desired page
	 * 
	 */

	public PhysicalPage store(VirtualAddress addr) {
		// reset Status & tlbEvictionAddress
		initStatus(addr);

		VirtualPage pageNum = addr.getPage();
		PhysicalPage pageFrame = null;
		PageTableEntry theEntry = pages.get(pageNum);
		if((theEntry == null) ||
				(theEntry.isSwapped()) ||
				!(theEntry.isValid())){
			if(pageCount < maxPages){
				pageFrame = new PhysicalPage(pageCount++);
				stat.setAccess(Access.CLEANMISS);
			} else {
				PageTableEntry oldEntry = getLRU();
				pageFrame = oldEntry.getPhysical();
				if(oldEntry.isValid() && oldEntry.isDirty()){
					stat.setDirty();
					this.dirtyMisses++;
					oldEntry.setSwapped(true);
				} else {
					stat.setClean();
					this.cleanMisses++;
				}
				oldEntry.setValid(false); // invalidate old entry
				this.tlbEvictionAddress = oldEntry.getTag();
			}
			theEntry = new PageTableEntry(pageNum,pageFrame);
			theEntry.setValid(true);
			pages.put(pageNum,theEntry);
		} else {
			this.hits++;
		}
		theEntry.setDirty(true); // Store is happening, make it dirty
		setMRU(theEntry);
		return theEntry.getPhysical();
	}

	public Status getStatus(){
		return stat;
	}

	public long getCleanMisses(){
		return this.cleanMisses;
	}

	public long getDirtyMisses(){
		return this.dirtyMisses;
	}

	public String getCounts(){
		long accesses = this.hits + this.cleanMisses + this.dirtyMisses;
		return accesses+"\t"+this.hits+"\t"+this.cleanMisses+"\t"+this.dirtyMisses;
	}
}
